@@ -29,7 +29,7 @@ class GZipCompression(Compression):
         return "gzip"
 
     def compress(self, data: bytes | bytearray) -> bytes:
-        return gzip.compress(data)
+        return gzip.compress(data, compresslevel=6)
 
     def decompress(self, data: bytes | bytearray) -> bytes:
         return gzip.decompress(data)
@@ -45,7 +45,7 @@ try:
             return "br"
 
         def compress(self, data: bytes | bytearray) -> bytes:
-            return brotli.compress(data)
+            return brotli.compress(data, quality=3)
 
         def decompress(self, data: bytes | bytearray) -> bytes:
             return brotli.decompress(data)
@@ -91,6 +91,10 @@ class IdentityCompression(Compression):
 _identity = IdentityCompression()
 _compressions["identity"] = _identity
 
+# Preferred compression names for Accept-Encoding header, in order of preference.
+# Excludes 'identity' since it's an implicit fallback.
+DEFAULT_ACCEPT_ENCODING_COMPRESSIONS = ("gzip", "br", "zstd")
+
 
 def get_compression(name: str) -> Compression | None:
     return _compressions.get(name.lower())
@@ -101,6 +105,17 @@ def get_available_compressions() -> KeysView:
     return _compressions.keys()
 
 
+def get_accept_encoding() -> str:
+    """Returns Accept-Encoding header value with available compressions in preference order.
+
+    This excludes 'identity' since it's an implicit fallback, and returns
+    only compressions that are actually available (i.e., their dependencies are installed).
+    """
+    return ", ".join(
+        name for name in DEFAULT_ACCEPT_ENCODING_COMPRESSIONS if name in _compressions
+    )
+
+
 def negotiate_compression(accept_encoding: str) -> Compression:
     for accept in accept_encoding.split(","):
         compression = _compressions.get(accept.strip())
